--[[
Big Eye Early Warning Radar script, moose based

new implementation by quelcertoleo
original EWRS script updated by Apple 30/11/2023
original EWRS concept by Steggles - https://github.com/Bob7heBuilder/EWRS

changelog
0.0.5 - Added Speed cutoff to filter slow movers, Checked for client to be reported to being alive
0.0.4 - remove requirement for ZONE_POLYGON objects defining airspaces
0.0.3 - added GCI and 12-hour report format
0.0.2 - added friendly picture and bogey dope
0.0.1 - initial mvp, supports player preferences and reports contacts on schedule
]]
-- CONSTANTS *************************************************************************************

EWR = {}
EWR.version = "0.0.5"
EWR.debug = false

_SETTINGS:SetPlayerMenuOff()
_SETTINGS:SetA2G_MGRS()
_SETTINGS:SetMGRS_Accuracy(4)

--[[
{
k:groupId
v:
[
    clientName,
    playerCoalition,
    reportReference,
    reportCardinalHeading,
    reportFrequency,
    reportEnabled,
    reportUnits,
    playerReportScheduler,
    isGciTaskingAvailable,
    playerGciScheduler,
    gciReportScheduler
]
}
]]
EWR.playerData = {}

EWR.radioMenusAdded = {}

-- Sets of detected units
EWR.blueDetectedUnits = SET_UNIT:New()
EWR.redDetectedUnits = SET_UNIT:New()

-- Constants
EWR.blueCoalition = "blue"
EWR.redCoalition = "red"
EWR.minThreatLevel = 0
EWR.maxThreatLevel = 10
EWR.referenceSelf = "self"
EWR.referenceBulls = "bulls"
EWR.unitInternationalSystem = "isu"
EWR.unitImperial = "imp"
EWR.SpeedCutoff = 100

-- Report timings and settings
EWR.displayMessageTime = 10
EWR.displayMessageTimeShort = 3
EWR.maxReportUnits = 5
EWR.highReportFrequencyPreference = 15 -- can't be less than EWRS.displayMessageTime
EWR.defaultReportFrequencyPreference = 30
EWR.lowReportFrequencyPreference = 60
EWR.gciReportFrequency = 5
EWR.defaultEnableReportPreference = true

-- GCI
-- {k:playerGroupId, v:[assignedUnit]}
EWR.gciAssignedUnits = {}
EWR.gciMaximumRange = 100000 -- meters

-- DETECTION Blue
EWR.blueHasDLINK = true
EWR.blueHasRWR = true
EWR.blueHasIRST = false
-- DETECTION Red
EWR.redHasDLINK = false
EWR.redHasRWR = true
EWR.redHasIRST = true

-- prefixes used in mission editor for EWR units
EWR.blueCoalitionRadarPrefixes = { "AWACS", "Blue Naval", "Blue SAM", "Blue EWR" }
EWR.redCoalitionRadarPrefixes = { "EWR-1", "EWR-2", "EWR-3", "EWR-8" }

-- in case of respawned units, use keepUnitNames so that they can be added to the filter dinamically
blueDetectionGroup = SET_GROUP:New():FilterPrefixes(EWR.blueCoalitionRadarPrefixes):FilterStart()
blueDetection = DETECTION_UNITS:New(blueDetectionGroup, 1000):InitDetectRadar(true)
blueDetection:InitDetectDLINK(EWR.blueHasDLINK)
blueDetection:InitDetectRWR(EWR.blueHasRWR)
blueDetection:InitDetectIRST(EWR.blueHasIRST)
blueDetection:FilterCategories({ Unit.Category.AIRPLANE, Unit.Category.HELICOPTER })
blueDetection:Start()

redDetectionGroup = SET_GROUP:New():FilterPrefixes(EWR.redCoalitionRadarPrefixes):FilterStart()
redDetection = DETECTION_UNITS:New(redDetectionGroup, 1000):InitDetectRadar(true)
redDetection:InitDetectDLINK(EWR.redHasDLINK)
redDetection:InitDetectRWR(EWR.redHasRWR)
redDetection:InitDetectIRST(EWR.redHasIRST)
redDetection:FilterCategories({ Unit.Category.AIRPLANE, Unit.Category.HELICOPTER })
redDetection:Start()

-- UTILS / MISC (yeah, misc and utils where stuff you don't know the concern goes) ***

-- @param #DETECTION_BASE.DetectedItem
-- @return #Wrapper.Unit#UNIT the detected unit
function EWR.detectedItemToWrapperUnit(detectedItem)
    local detectedSet = detectedItem.Set
    local detectedUnit = detectedSet:GetFirst()
    return detectedUnit
end

-- @param DCS#Vec2 Vec2 objects to calculate distance
-- @return #number distance in meters
function EWR.getDistance(object1Vec2, object2Vec2)
    local xDiff = object1Vec2.x - object2Vec2.x
    local yDiff = object1Vec2.y - object2Vec2.y
    local distanceInMeters = math.sqrt(xDiff * xDiff + yDiff * yDiff)
    return distanceInMeters
end

-- @param DCS#Vec2 objectFromVec2 the vertex
-- @param DCS#Vec2 objectToVec2 the destination
-- @return #number angle in degrees
function EWR.getAbsoluteBearing(objectFromVec2, objectToVec2)
    -- Calculate the relative bearing
    local bearing = math.atan2(objectToVec2.y - objectFromVec2.y, objectToVec2.x - objectFromVec2.x)

    -- Convert the bearing to degrees
    bearing = math.deg(bearing)

    -- Make sure the bearing is positive
    bearing = (bearing + 360) % 360
    return UTILS.Round(bearing, -1)
end

-- @param DCS#Vec2 objectFromVec2 the vertex
-- @param DCS#Vec2 objectToVec2 the destination
-- @return #number angle in degrees
function EWR.getRelativeBearing(objectFromVec2, objectToVec2, objectFromHeading)
    -- Calculate the relative bearing
    local bearing = math.atan2(objectToVec2.y - objectFromVec2.y, objectToVec2.x - objectFromVec2.x)
    env.info("bearing " .. bearing)
    -- Convert the bearing to degrees
    bearing = math.deg(bearing)
    env.info("bearing deg " .. bearing)

    -- Make sure the bearing is positive
    bearing = (bearing - objectFromHeading + 360) % 360
    env.info("bearing positive " .. bearing)

    return UTILS.Round(bearing, -1)
end

--- Heading Degrees (0-360) to 12-hour format
-- @param #number bearing is the bearing in degrees
-- @return #string in 12-hour format, e.g. "11 o'clock"
function EWR.getBearingToClockFormat(bearing)

    local bearingClock = math.floor(bearing / 30)

    if bearingClock == 0 then
        bearingClock = 12
    end

    return string.format("%d o'clock", bearingClock)
end

-- @param #string clientName
-- @param #string playerCoalition EWR.blueCoalition / EWR.redCoalition
-- @param clientReferencePreference EWR.playerEwrPreferences["reportReference"]
-- @return DCS#Vec2 Vec2 of reference (self / bulls)
function EWR.getReferenceVec2ForClient(clientName, playerCoalition, clientReferencePreference)
    local referenceVec2
    if clientReferencePreference == EWR.referenceBulls and playerCoalition == EWR.blueCoalition then
        local blueBullsCoordinate = COORDINATE.GetBullseyeCoordinate(coalition.side.BLUE)
        referenceVec2 = blueBullsCoordinate:GetVec2()
    elseif clientReferencePreference == EWR.referenceBulls and playerCoalition == EWR.redCoalition then
        local redBullsCoordinate = COORDINATE.GetBullseyeCoordinate(coalition.side.RED)
        referenceVec2 = redBullsCoordinate:GetVec2()
    else
        local clientUnit = UNIT:FindByName(clientName)
        referenceVec2 = clientUnit:GetVec2()
    end
    return referenceVec2
end

-- @param #SET_UNIT setUnit, in our case the SET_UNITS that contain detected units
-- @param DCS#Vec2 referenceVec2 Vec2 of reference point
-- @return #Wrapper.Unit#UNIT the closest unit
function EWR.getClosestUnitInSetTo(setUnit, referenceVec2)
    if setUnit:CountAlive() == 0 then
        return nil
    else
        local closestUnit = setUnit:GetFirst()
        local closestUnitDistance = EWR.getDistance(closestUnit:GetVec2(), referenceVec2)
        setUnit:ForEachUnit(
                function(unit)
                    local detectedUnitVec2 = unit:GetVec2()
                    local distanceFromDetectedUnit = EWR.getDistance(detectedUnitVec2, referenceVec2)
                    if distanceFromDetectedUnit < closestUnitDistance then
                        local speed = unit:GetVelocityKNOTS()
                        if speed > EWR.SpeedCutoff then
                          closestUnitDistance = distanceFromDetectedUnit
                          closestUnit = unit
                        end
                    end
                end
        )
        return closestUnit
    end
end

-- @param #SET_UNIT setUnit, in our case the SET_UNITS that contain detected units
-- @return #Wrapper.Unit#UNIT the highest threat in the coalition airspace
function EWR.assignHighestThreatInAirspaceToPlayer(groupId)

    local playerPreferences = EWR.playerData[groupId]
    local playerCoalition = playerPreferences["playerCoalition"]
    local assignedTargetUnit
    if playerCoalition == EWR.blueCoalition then
        assignedTargetUnit = EWR.assignFromSetInAirspace(groupId, EWR.blueDetectedUnits)
    else
        assignedTargetUnit = EWR.assignFromSetInAirspace(groupId, EWR.redDetectedUnits)
    end
    if assignedTargetUnit ~= nil then
        trigger.action.outTextForGroup(groupId, string.format("Assigned target %s", assignedTargetUnit:GetNatoReportingName()), 10, false)
        local gciAssignTargetScheduler = EWR.playerData[groupId]["playerGciScheduler"]
        gciAssignTargetScheduler:Stop()
        local gciTargetReportScheduler = EWR.getGciTargetReportSchedulerForPlayerGroup(groupId, assignedTargetUnit)
        gciTargetReportScheduler:Start()
    else
        trigger.action.outTextForGroup(groupId, "Scanning contacts for GCI tasking", 10, false)
    end
end

-- @param #number groupId of client group
-- @param #SET_UNIT setUnit, in our case the SET_UNITS that contain detected units
-- @param #ZONE_POLYGON the airspace zone
-- @return #Wrapper.Unit#UNIT the assigned target unit
function EWR.assignFromSetInAirspace(groupId, detectedUnitSet)

    local clientName = EWR.playerData[groupId]["clientName"]
    local clientUnit = UNIT:FindByName(clientName)
    local clientVec2 = clientUnit:GetVec2()
    local firstUnitDistance = EWR.gciMaximumRange

    detectedUnitSet:ForEachUnitPerThreatLevel(EWR.maxThreatLevel, EWR.minThreatLevel,
            function(unit)
                if unit:IsAlive() and EWR.gciAssignedUnits[groupId] == nil then
                    local unitVec2 = unit:GetVec2()
                    local unitDistance = EWR.getDistance(clientVec2, unitVec2)
                    if unitDistance < firstUnitDistance then
                        EWR.gciAssignedUnits[groupId] = unit
                    end
                end
            end
    )
    return EWR.gciAssignedUnits[groupId]
end

-- PREFERENCES ***

-- @param #number groupId of client group
-- @param #string referencePreference EWRS.playerEwrPreferences[groupId]["reportReference"]
function EWR.setReferencePreference(groupId, referencePreference)
    if EWR.debug then
        env.info(string.format("EWRS group %s set preference refecence to %s", groupId, referencePreference))
    end
    local playerPreferences = EWR.playerData[groupId]
    playerPreferences["reportReference"] = referencePreference
    local newReportScheduler = EWR.getReportSchedulerForPlayerGroup(groupId)
    EWR.replacePlayerReportSchedulerWith(groupId, newReportScheduler)
    trigger.action.outTextForGroup(groupId, string.format("Reference to %s", referencePreference), 10, false)
end

-- @param #number groupId of client group
-- @param #boolean isHeadingCardinal EWRS.playerEwrPreferences[groupId]["reportCardinalHeading"]
function EWR.setCardinalHeadingPreference(groupId, isHeadingCardinal)
    if EWR.debug then
        env.info(string.format("EWRS group %s set heading refecence to %s", groupId, tostring(isHeadingCardinal)))
    end
    local playerPreferences = EWR.playerData[groupId]
    playerPreferences["reportCardinalHeading"] = isHeadingCardinal
    local newReportScheduler = EWR.getReportSchedulerForPlayerGroup(groupId)
    EWR.replacePlayerReportSchedulerWith(groupId, newReportScheduler)
    if isHeadingCardinal then
        trigger.action.outTextForGroup(groupId, "Heading to cardinal", 10, false)
    else
        trigger.action.outTextForGroup(groupId, "Heading to numerical", 10, false)
    end
end

-- @param #number groupId of client group
-- @param #number unitPreference EWRS.playerEwrPreferences[groupId]["reportUnits"]
function EWR.setUnitPreference(groupId, unitPreference)
    if EWR.debug then
        env.info(string.format("EWR group %s set unit preference to %s", groupId, unitPreference))
    end
    local playerPreferences = EWR.playerData[groupId]
    playerPreferences["reportUnits"] = unitPreference
    local newReportScheduler = EWR.getReportSchedulerForPlayerGroup(groupId)
    EWR.replacePlayerReportSchedulerWith(groupId, newReportScheduler)
    trigger.action.outTextForGroup(groupId, string.format("Unit to %s", unitPreference), 10, false)
end

-- @param #number groupId of client group
-- @param #string reportFrequencyPreference EWRS.playerEwrPreferences[groupId]["reportFrequency"]
function EWR.setReportFrequencyPreference(groupId, reportFrequencyPreference)
    if EWR.debug then
        env.info(string.format("EWR group %s set preference report frequency to %2d", groupId, reportFrequencyPreference))
    end
    local playerPreferences = EWR.playerData[groupId]
    playerPreferences["reportFrequency"] = reportFrequencyPreference
    local newReportScheduler = EWR.getReportSchedulerForPlayerGroup(groupId)
    EWR.replacePlayerReportSchedulerWith(groupId, newReportScheduler)
    trigger.action.outTextForGroup(groupId, string.format("Frequency to %2ds", reportFrequencyPreference), 10, false)
end

-- @param #number groupId of client group
-- @param #boolean reportEnabledPreference boolean, will start or stop player scheduler in EWR.playerEwrPreferences[groupId]["playerReportScheduler"]
function EWR.setReportEnabledPreference(groupId, reportEnabledPreference)
    if EWR.debug then
        env.info(string.format("EWR group %s set preference report enable %s", groupId, tostring(reportEnabledPreference)))
    end
    local playerReportScheduler = EWR.playerData[groupId]["playerReportScheduler"]
    if reportEnabledPreference then
        trigger.action.outTextForGroup(groupId, "Enable contact report", 10, false)
        playerReportScheduler:Start()
    else
        trigger.action.outTextForGroup(groupId, "Disable contact report", 10, false)
        playerReportScheduler:Stop()
    end
    EWR.playerData[groupId]["reportEnabled"] = reportEnabledPreference
end

-- @param #number groupId of client group
-- @param #boolean reportEnabledPreference boolean, will start or stop player scheduler in EWR.playerEwrPreferences[groupId]["playerReportScheduler"]
function EWR.setGciTaskingAvailablePreference(groupId, gciEnabledPreference)
    if EWR.debug then
        env.info(string.format("EWR group %s set preference GCI available %s", groupId, tostring(gciEnabledPreference)))
    end
    local playerReportScheduler = EWR.playerData[groupId]["playerReportScheduler"]
    local playerGciAssignTargetScheduler = EWR.playerData[groupId]["playerGciScheduler"]
    local playerGciReportScheduler = EWR.playerData[groupId]["gciReportScheduler"]
    if gciEnabledPreference then
        trigger.action.outTextForGroup(groupId, "Enable GCI tasking, stopping reports", 10, false)
        playerReportScheduler:Stop()
        playerGciAssignTargetScheduler:Start()
    else
        trigger.action.outTextForGroup(groupId, "Disable GCI tasking", 10, false)
        playerGciAssignTargetScheduler:Stop()
        playerGciReportScheduler:Stop()
        EWR.gciAssignedUnits[groupId] = nil
    end
    EWR.playerData[groupId]["isGciTaskingAvailable"] = gciEnabledPreference
end

-- SCHEDULERS ***

-- @param #number groupId of client group
-- @param #table playerPreferences EWR.playerEwrPreferences[groupId]
-- @return #SCHEDULER
function EWR.getReportSchedulerForPlayerGroup(groupId)
    local playerPreferences = EWR.playerData[groupId]
    local playerReportScheduler = SCHEDULER:New(nil,
            function()
                local isOnDemandReport = false
                EWR.printBogeyReport(groupId, playerPreferences["clientName"], playerPreferences["playerCoalition"], isOnDemandReport)
            end, {}, 0, playerPreferences["reportFrequency"], 0)
    return playerReportScheduler
end

-- @param #number groupId of client group
-- @param #table playerPreferences EWR.playerEwrPreferences[groupId]
-- @return #SCHEDULER
function EWR.getGciSchedulerForPlayerGroup(groupId)
    local gciAssignTargetFrequency = 60
    local playerGciScheduler = SCHEDULER:New(nil,
            function()
                EWR.assignHighestThreatInAirspaceToPlayer(groupId)
            end, {}, 0, gciAssignTargetFrequency, 0)
    EWR.playerData[groupId]["playerGciScheduler"] = playerGciScheduler
    return playerGciScheduler
end

-- @param #number groupId of client group
function EWR.getGciTargetReportSchedulerForPlayerGroup(groupId, assignedTargetUnit)
    local playerGciScheduler = SCHEDULER:New(nil,
            function()
                EWR.printGciReport(groupId, assignedTargetUnit)
            end, {}, 0, EWR.displayMessageTimeShort, 0)
    EWR.playerData[groupId]["gciReportScheduler"] = playerGciScheduler
    return playerGciScheduler
end

-- @param #number groupId of client group
-- @param #SCHEDULER the new scheduler with modified player preferences, stop the old one and start a new scheduler
function EWR.replacePlayerReportSchedulerWith(groupId, newReportScheduler)
    local playerPreferences = EWR.playerData[groupId]
    local playerReportScheduler = playerPreferences["playerReportScheduler"]
    playerReportScheduler:Stop()
    playerPreferences["playerReportScheduler"] = newReportScheduler
    playerReportScheduler = playerPreferences["playerReportScheduler"]
    playerReportScheduler:Start()
end

-- DETECTION CALLBACKS ***
function blueDetection:OnAfterDetectedItem(From, Event, To, DetectedItem)
    local detectedUnit = EWR.detectedItemToWrapperUnit(DetectedItem)
    EWR.blueDetectedUnits:AddUnit(detectedUnit)
    if EWR.debug then
        env.info(string.format("EWR BLUE detected %s", detectedUnit:GetName()))
    end
end

function redDetection:OnAfterDetectedItem(From, Event, To, DetectedItem)
    local detectedUnit = EWR.detectedItemToWrapperUnit(DetectedItem)
    EWR.redDetectedUnits:AddUnit(detectedUnit)
    if EWR.debug then
        env.info(string.format("EWR RED detected %s", detectedUnit:GetName()))
    end
end

-- REPORTS ***

-- @param #number playerGroupId of client group
-- @param #Wrapper.Unit#UNIT assignedTargetUnit is the unit to generate report of
function EWR.printGciReport(groupId, assignedTargetUnit)
    if EWR.debug then
        env.info(string.format("EWR GCI report for group %3d, %s", groupId, assignedTargetUnit:GetName()))
    end

    local clientName = EWR.playerData[groupId]["clientName"]
    local clientUnit = UNIT:FindByName(clientName)
    local clientVec2 = clientUnit:GetVec2()

    local clientPreferences = EWR.playerData[groupId]
    if assignedTargetUnit ~= nil and assignedTargetUnit:IsAlive() then
        local assignedTargetUnitVec2 = assignedTargetUnit:GetVec2()
        local assignedTargetUnitDistance = EWR.getDistance(clientVec2, assignedTargetUnitVec2)
        local isAltitudeFromGround = false
        local clientAltitude = clientUnit:GetAltitude(isAltitudeFromGround)
        local assignedTargetAltitude = assignedTargetUnit:GetAltitude(isAltitudeFromGround)

        if (assignedTargetUnitDistance <= UTILS.NMToMeters(5)) then

            local unitDistanceToString
            if clientPreferences["reportUnits"] == EWR.unitImperial then
                unitDistanceToString = string.format("%03dnm", UTILS.Round(UTILS.MetersToNM(assignedTargetUnitDistance), 0))
            else
                unitDistanceToString = string.format("%03dkm", UTILS.Round(assignedTargetUnitDistance / 1000, 0))
            end

            local bearingInDegrees = EWR.getRelativeBearing(clientVec2, assignedTargetUnitVec2, clientUnit:GetHeading())
            local bearingInClockFormat = EWR.getBearingToClockFormat(bearingInDegrees)

            local relativeAltitude = ""
            if clientAltitude - assignedTargetAltitude > 1000 then
                relativeAltitude = "low"
            elseif clientAltitude - assignedTargetAltitude < -1000 then
                relativeAltitude = "high"
            end

            local unitTextReport = string.format("%s | %s %s", unitDistanceToString, bearingInClockFormat, relativeAltitude)
            trigger.action.outTextForGroup(groupId, unitTextReport, EWR.displayMessageTimeShort, false)
        else
            local altitudeString
            if assignedTargetAltitude == nil then
                altitudeString = string.format("%8s", "???")
            elseif assignedTargetAltitude < 350 then
                altitudeString = string.format("%8s", "low")
            else
                if clientPreferences["reportUnits"] == EWR.unitImperial then
                    altitudeString = string.format("%05dft", UTILS.Round(UTILS.MetersToFeet(assignedTargetAltitude), -3))
                else
                    altitudeString = string.format("%05dmt", UTILS.Round(assignedTargetAltitude, -2))
                end
            end

            local bearing = EWR.getAbsoluteBearing(clientVec2, assignedTargetUnitVec2)
            local unitDistance = EWR.getDistance(clientVec2, assignedTargetUnitVec2)
            local unitDistanceToString
            if clientPreferences["reportUnits"] == EWR.unitImperial then
                unitDistanceToString = string.format("%03dnm", UTILS.Round(UTILS.MetersToNM(unitDistance), 0))
            else
                unitDistanceToString = string.format("%03dkm", UTILS.Round(unitDistance / 1000, 0))
            end
            local bearingDistanceString = string.format("%03d° for %s", bearing, unitDistanceToString)

            local unitTextReport = string.format(
                    "%-10s | %s | %s",
                    assignedTargetUnit:GetNatoReportingName(),
                    bearingDistanceString,
                    altitudeString)

            trigger.action.outTextForGroup(groupId, unitTextReport, EWR.displayMessageTimeShort, false)
        end
    else
        trigger.action.outTextForGroup(groupId, "Target destroyed, good job", EWR.displayMessageTime, false)
        EWR.setGciTaskingAvailablePreference(groupId, false)
    end
end

-- @param #number groupId of client group
-- @param #string clientName string the name of the client group (and unit) as defined in Mission Editor
-- @param #string EWR.blueCoalition / EWR.redCoalition
function EWR.printBogeyDope(playerGroupId, clientName, playerCoalition)
    if EWR.debug then
        env.info(string.format("EWR bogey dope for group %s, %s, %s", playerGroupId, clientName, playerCoalition))
    end

    local clientUnit = UNIT:FindByName(clientName)
    local clientVec2 = clientUnit:GetVec2()
    local closestUnit
    if playerCoalition == EWR.blueCoalition then
        closestUnit = EWR.getClosestUnitInSetTo(EWR.blueDetectedUnits, clientVec2)
    else
        closestUnit = EWR.getClosestUnitInSetTo(EWR.redDetectedUnits, clientVec2)
    end

    if closestUnit ~= nil and closestUnit:IsAlive() then
        if EWR.debug then
            env.info(string.format("EWR bogey closest for group %s is %s", clientName, closestUnit:GetName()))
        end
        local clientPreferences = EWR.playerData[playerGroupId]
        local unitTextReport = EWR.getTextReportForUnit(closestUnit, clientPreferences, clientVec2)
        local contactReport = REPORT:New("Bogey dope")
        contactReport:Add(unitTextReport)
        trigger.action.outTextForGroup(playerGroupId, contactReport:Text(), EWR.displayMessageTime, false)
    else
        trigger.action.outTextForGroup(playerGroupId, "clean", EWR.displayMessageTime, false)
    end
end

-- @param #number groupId of client group
-- @param #string clientName string the name of the client group (and unit) as defined in Mission Editor
-- @param #string EWR.blueCoalition / EWR.redCoalition
-- @param #boolean isOnDemand boolean used to print "clean" when report has been explictly requested and no units around, if it's scheduled report then stay quiet
function EWR.printBogeyReport(groupId, clientName, playerCoalition, isOnDemand)
    if EWR.debug then
        env.info(string.format("EWR report for group %s", groupId))
    end
    local client = CLIENT:FindByName(clientName)
    if not client:IsAlive() then return end
    if isOnDemand then
        if playerCoalition == EWR.blueCoalition then
            if EWR.blueDetectedUnits:Count() == 0 then
                trigger.action.outTextForGroup(groupId, "clean", EWR.displayMessageTime, false)
                return
            end
        else
            if EWR.redDetectedUnits:Count() == 0 then
                trigger.action.outTextForGroup(groupId, "clean", EWR.displayMessageTime, false)
                return
            end
        end
    end

    local clientPreferences = EWR.playerData[groupId]
    local clientReferencePreference = clientPreferences["reportReference"]

    local contactReport = REPORT:New(string.format("Reference %s", clientReferencePreference))
    local referenceVec2 = EWR.getReferenceVec2ForClient(clientName, playerCoalition, clientReferencePreference)

    local reportUnitCounter = 1
    if playerCoalition == EWR.blueCoalition then
        EWR.blueDetectedUnits:ForEachUnitPerThreatLevel(EWR.maxThreatLevel, EWR.minThreatLevel,
                function(unit)
                    if reportUnitCounter <= EWR.maxReportUnits and unit ~= nil and unit:IsAlive() and unit:GetVelocityKNOTS() > EWR.SpeedCutoff then
                        local unitTextReport = EWR.getTextReportForUnit(unit, clientPreferences, referenceVec2)
                        contactReport:Add(unitTextReport)
                        reportUnitCounter = reportUnitCounter + 1
                    end
                end
        )
    else
        EWR.redDetectedUnits:ForEachUnitPerThreatLevel(EWR.maxThreatLevel, EWR.minThreatLevel,
                function(unit)
                    if reportUnitCounter <= EWR.maxReportUnits and unit ~= nil and unit:IsAlive() and unit:GetVelocityKNOTS() > EWR.SpeedCutoff then
                        local unitTextReport = EWR.getTextReportForUnit(unit, clientPreferences, referenceVec2)
                        contactReport:Add(unitTextReport)
                        reportUnitCounter = reportUnitCounter + 1
                    end
                end
        )
    end

    if playerCoalition == EWR.blueCoalition then
        if EWR.blueDetectedUnits:Count() > 0 then
            trigger.action.outTextForGroup(groupId, contactReport:Text(), EWR.displayMessageTime, false)
        end
    else
        if EWR.redDetectedUnits:Count() > 0 then
            trigger.action.outTextForGroup(groupId, contactReport:Text(), EWR.displayMessageTime, false)
        end
    end
end

-- @param #number groupId of client group
-- @param #string clientName string the name of the client group (and unit) as defined in Mission Editor
-- @param #string playerCoalition EWR.redCoalition / EWR.blueCoalition
function EWR.printFriendliesReport(playerGroupId, clientName, playerCoalition)
    if EWR.debug then
        env.info(string.format("EWR find friendlies for %3d %s", playerGroupId, clientName))
    end
    -- cannot use { Unit.Category.AIRPLANE, Unit.Category.HELICOPTER } here, inconsistent API with others
    local friendlyUnitSet = SET_UNIT
            :New()
            :FilterActive(true)
            :FilterCoalitions(playerCoalition)
            :FilterCategories({ "plane", "helicopter" })
            :FilterOnce()
    -- remove client unit from set
    friendlyUnitSet:RemoveUnitsByName(clientName)

    if friendlyUnitSet:CountAlive() == 0 then
        trigger.action.outTextForGroup(playerGroupId, "no friendlies", EWR.displayMessageTime, false)
        return
    end

    local clientUnit = UNIT:FindByName(clientName)
    local clientUnitVec2 = clientUnit:GetVec2()

    local friendliesTable = {}
    local counter = 1
    friendlyUnitSet:ForEachUnit(
            function(unit)
                local speed = unit:GetVelocityKNOTS()
                if speed > EWR.SpeedCutoff then
                  friendliesTable[counter] = {}
                  friendliesTable[counter].unitName = unit:GetName()
                  local unitDistance = EWR.getDistance(clientUnitVec2, unit:GetVec2())
                  friendliesTable[counter].distance = unitDistance
                  counter = counter + 1
                end
            end
    )

    local function sortTableByDistance(unit1, unit2)
        return unit2.distance > unit1.distance
    end
    table.sort(friendliesTable, sortTableByDistance)

    local friendlyReport = REPORT:New("Friendly picture")
    local clientPreferences = EWR.playerData[playerGroupId]
    for i = 1, #friendliesTable do
        if i == EWR.maxReportUnits + 1 then
            break
        end
        local friendlyUnit = UNIT:FindByName(friendliesTable[i].unitName)
        if friendlyUnit ~= nil and friendlyUnit:IsAlive() then
            local friendlyTextReport = EWR.getTextReportForUnit(friendlyUnit, clientPreferences, clientUnitVec2)
            friendlyReport:Add(friendlyTextReport)
        end
    end
    trigger.action.outTextForGroup(playerGroupId, friendlyReport:Text(), EWR.displayMessageTime, false)
end

-- @param #Wrapper.Unit#UNIT unit is the unit to generate report of
-- @param #table clientPreferences EWRS.playerEwrPreferences[groupId]
-- @param #DCS#Vec2 referenceVec2 Vec2 used to calculate relative distance, heading, tracking
-- @return #string with report entry for @param unit
function EWR.getTextReportForUnit(unit, clientPreferences, referenceVec2)
    if EWR.debug then
        env.info(string.format("EWR prepare text report for %s", unit:GetName()))
    end

    -- TODO extract to function?
    local isAltitudeFromGround = false
    local unitAltitude = unit:GetAltitude(isAltitudeFromGround)
    local altitudeString
    if unitAltitude == nil then
        altitudeString = string.format("%8s", "???")
    elseif unitAltitude < 1000 then
        altitudeString = string.format("%8s", "low")
    else
        if clientPreferences["reportUnits"] == EWR.unitImperial then
            altitudeString = string.format("%05dft", UTILS.Round(UTILS.MetersToFeet(unitAltitude), -3))
        else
            altitudeString = string.format("%05dmt", UTILS.Round(unitAltitude, -2))
        end
    end

    local speedString
    -- Sort out slow movers
    local isslow = false
    local speed = unit:GetVelocityKNOTS()
    if speed < EWR.SpeedCutoff then isslow = true end
    
    if clientPreferences["reportUnits"] == EWR.unitImperial then
        speedString = string.format("%03dkts", unit:GetVelocityKNOTS())
    else
        speedString = string.format("%03dkph", unit:GetVelocityKMH())
    end

    -- TODO extract to function?
    local unitVec2 = unit:GetVec2()
    local bearing = EWR.getAbsoluteBearing(referenceVec2, unitVec2)
    local unitDistance = EWR.getDistance(unitVec2, referenceVec2)
    local unitDistanceToString
    if clientPreferences["reportUnits"] == EWR.unitImperial then
        unitDistanceToString = string.format("%03dnm", UTILS.Round(UTILS.MetersToNM(unitDistance), 0))
    else
        unitDistanceToString = string.format("%03dkm", UTILS.Round(unitDistance / 1000, 0))
    end
    local bearingDistanceString = string.format("%03d° for %s", bearing, unitDistanceToString)

    -- TODO extract to function?
    local heading = UTILS.Round(unit:GetHeading(), -1)
    if clientPreferences["reportCardinalHeading"] then
        heading = UTILS.BearingToCardinal(heading)
    else
        heading = string.format("%03d°", heading)
    end

    local unitTextReport = string.format(
            "%-10s | %s | %s | %s | %s",
            unit:GetNatoReportingName(),
            bearingDistanceString,
            speedString,
            altitudeString,
            heading)

    return unitTextReport, isslow
end

-- SETUP CLIENTS ***
function EWR.setupClients()
    local bluePlayerClients = SET_CLIENT:New():FilterCoalitions(EWR.blueCoalition):FilterActive(true):FilterOnce()
    bluePlayerClients:ForEachClient(
            function(client)
                local playerGroupId = client:GetClientGroupID()
                if EWR.radioMenusAdded[playerGroupId] == nil then
                    EWR.setupClientsHelper(playerGroupId, client, EWR.blueCoalition)
                end
            end
    )
    local redPlayerClients = SET_CLIENT:New():FilterCoalitions(EWR.redCoalition):FilterActive(true):FilterOnce()
    redPlayerClients:ForEachClient(
            function(client)
                local playerGroupId = client:GetClientGroupID()
                if EWR.radioMenusAdded[playerGroupId] == nil then
                    EWR.setupClientsHelper(playerGroupId, client, EWR.redCoalition)
                end
            end
    )
end

function EWR.setupClientsHelper(playerGroupId, client, playerCoalition)
    if EWR.debug then
        env.info(string.format("EWR client setup for group %s %s", playerGroupId, client:GetName()))
    end

    local reportUnit
    if playerCoalition == EWR.redCoalition then
        reportUnit = EWR.unitInternationalSystem
    else
        reportUnit = EWR.unitImperial
    end

    EWR.playerData[playerGroupId] = {
        groupId = playerGroupId,
        clientName = client:GetName(),
        playerCoalition = playerCoalition,
        reportReference = EWR.referenceSelf,
        reportCardinalHeading = true,
        reportFrequency = EWR.defaultReportFrequencyPreference,
        reportEnabled = EWR.defaultEnableReportPreference,
        reportUnits = reportUnit,
        playerReportScheduler = nil,
        isGciTaskingAvailable = false,
        playerGciScheduler = nil,
        gciReportScheduler = nil
    }
    EWR.playerData[playerGroupId]["playerReportScheduler"] = EWR.getReportSchedulerForPlayerGroup(playerGroupId)
    EWR.playerData[playerGroupId]["playerGciScheduler"] = EWR.getGciSchedulerForPlayerGroup(playerGroupId)
    EWR.playerData[playerGroupId]["playerGciScheduler"]:Stop()

    local isOnDemandReport = true
    -- TODO remove playerCoalition from these calls, not needed
    -- TODO remove client:GetName() from these calls, not needed
    
    local MenuRoot = missionCommands.addSubMenuForGroup(playerGroupId, "BigEye EWR")
    
    missionCommands.addCommandForGroup(
            playerGroupId, "EWR Bogey Dope", MenuRoot, EWR.printBogeyDope, playerGroupId, client:GetName(), playerCoalition)
    missionCommands.addCommandForGroup(
            playerGroupId, "EWR Contact Report", MenuRoot, EWR.printBogeyReport, playerGroupId, client:GetName(), playerCoalition, isOnDemandReport)
    missionCommands.addCommandForGroup(
            playerGroupId, "EWR Friendly Picture", MenuRoot, EWR.printFriendliesReport, playerGroupId, client:GetName(), playerCoalition)

    local ewrsMenuRoot = missionCommands.addSubMenuForGroup(playerGroupId, "EWR Preferences", MenuRoot)

    local gciMenuRoot = missionCommands.addSubMenuForGroup(
            playerGroupId, "EWR GCI Tasking", ewrsMenuRoot)
    missionCommands.addCommandForGroup(
            playerGroupId, "available for tasking", gciMenuRoot, EWR.setGciTaskingAvailablePreference, playerGroupId, true)
    missionCommands.addCommandForGroup(
            playerGroupId, "tasking out", gciMenuRoot, EWR.setGciTaskingAvailablePreference, playerGroupId, false)
    
    local ewrsReportsMenuRoot = missionCommands.addSubMenuForGroup(
            playerGroupId, "Reports", ewrsMenuRoot)
    missionCommands.addCommandForGroup(
            playerGroupId, "enable contact reports", ewrsReportsMenuRoot, EWR.setReportEnabledPreference, playerGroupId, true)
    missionCommands.addCommandForGroup(
            playerGroupId, "disable contact reports", ewrsReportsMenuRoot, EWR.setReportEnabledPreference, playerGroupId, false)

    local ewrsReferenceMenuRoot = missionCommands.addSubMenuForGroup(
            playerGroupId, "Reference", ewrsMenuRoot)
    missionCommands.addCommandForGroup(
            playerGroupId, "bulls reference", ewrsReferenceMenuRoot, EWR.setReferencePreference, playerGroupId, EWR.referenceBulls)
    missionCommands.addCommandForGroup(
            playerGroupId, "self reference", ewrsReferenceMenuRoot, EWR.setReferencePreference, playerGroupId, EWR.referenceSelf)

    local ewrsHeadingMenuRoot = missionCommands.addSubMenuForGroup(
            playerGroupId, "Heading", ewrsMenuRoot)
    missionCommands.addCommandForGroup(
            playerGroupId, "cardinal heading", ewrsHeadingMenuRoot, EWR.setCardinalHeadingPreference, playerGroupId, true)
    missionCommands.addCommandForGroup(
            playerGroupId, "numerical heading", ewrsHeadingMenuRoot, EWR.setCardinalHeadingPreference, playerGroupId, false)

    local ewrsUnitsMenuRoot = missionCommands.addSubMenuForGroup(
            playerGroupId, "Units", ewrsMenuRoot)
    missionCommands.addCommandForGroup(
            playerGroupId, "unit imperial", ewrsUnitsMenuRoot, EWR.setUnitPreference, playerGroupId, EWR.unitImperial)
    missionCommands.addCommandForGroup(
            playerGroupId, "unit metric", ewrsUnitsMenuRoot, EWR.setUnitPreference, playerGroupId, EWR.unitInternationalSystem)

    local ewrsFrecuencyMenuRoot = missionCommands.addSubMenuForGroup(
            playerGroupId, "Frequency", ewrsMenuRoot)
    missionCommands.addCommandForGroup(
            playerGroupId, "high frequency reports", ewrsFrecuencyMenuRoot, EWR.setReportFrequencyPreference, playerGroupId, EWR.highReportFrequencyPreference)
    missionCommands.addCommandForGroup(
            playerGroupId, "default frequency reports", ewrsFrecuencyMenuRoot, EWR.setReportFrequencyPreference, playerGroupId, EWR.defaultReportFrequencyPreference)
    missionCommands.addCommandForGroup(
            playerGroupId, "low frequency reports", ewrsFrecuencyMenuRoot, EWR.setReportFrequencyPreference, playerGroupId, EWR.lowReportFrequencyPreference)

    EWR.radioMenusAdded[playerGroupId] = true
end

-- reschedule for users that change slot or coalition
setupClientScheduler = SCHEDULER:New(nil,
        function()
            EWR.setupClients()
        end, {}, 0, 30, 0)

-- cleanup schedulers of users that are dead or left the mission
cleanupScheduler = SCHEDULER:New(nil,
        function()
            for groupId, clientPreferences in pairs(EWR.playerData) do
                local clientName = clientPreferences["clientName"]
                local client = CLIENT:FindByName(clientName, nil, false)
                if client == nil or not client:IsPlayer() then
                    if EWR.debug then
                        env.info(string.format("EWR client clean up for %s", groupId))
                    end
                    local playerReportScheduler = EWR.getReportSchedulerForPlayerGroup(groupId)
                    playerReportScheduler:Stop()
                    local gciReportScheduler = EWR.getGciSchedulerForPlayerGroup(groupId)
                    gciReportScheduler:Stop()
                    env.info(string.format("EWR stopping scheduler for %s, %s", clientPreferences["groupId"], clientPreferences["clientName"]))
                    EWR.playerData[groupId] = nil
                    EWR.radioMenusAdded[groupId] = nil
                end
            end
        end, {}, 0, 60, 0)

env.info("Big Eye EWR " .. EWR.version .. " Running")
trigger.action.outText("Big Eye EWR " .. EWR.version .. " Running", 10)
